""" 
Classes
-------

Containers for figures, plots, etc.
"""
from __future__ import annotations

from typing import TYPE_CHECKING
import numpy as np
import pyqtgraph as pg
from accwidgets.graph import StaticPlotWidget
from accwidgets.graph.widgets.plotitem import ExViewBox
from accwidgets.graph.widgets.plotwidget import ExPlotWidgetProperties
from qtpy.QtCore import Signal, Qt

if TYPE_CHECKING:
    from pyqtgraph.GraphicsScene import mouseEvents
    from qtpy.QtWidgets import QGraphicsSceneMouseEvent

YAXES_WIDTH: int = 60

class ObservablePlotDataItem(pg.PlotDataItem):
    """A PlotDataItem that emits a signal when visibility changes."""
    visibilityChanged = Signal(bool)

    def setVisible(self, visible):
        super().setVisible(visible)
        self.visibilityChanged.emit(visible)


class DualPlotWidget(pg.LayoutWidget):
    """ A widget containing and handling two plots stacked vertically. """
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        pg.setConfigOptions(antialias=True)  # not sure if best place here

        self.top = PlotWidget()
        self.bottom = PlotWidget()

        self.addWidget(self.top, row=0, col=0)
        self.addWidget(self.bottom, row=1, col=0)

        # set margins, so the axes line up
        for plot in (self.top, self.bottom):
            plot.setContentsMargins(0, 0, 0, 0)
            plot.plotItem.getAxis("left").setWidth(YAXES_WIDTH)  # keep constant

    @property
    def plots(self) -> tuple[pg.PlotWidget, pg.PlotWidget]:
        return (self.top, self.bottom)

    def clear(self) -> None:    
        for plot in self.plots:
            plot.clear()
            plot.enableAutoRange()
    
    def set_connect_x(self, connect: bool) -> None:
        if connect:
            self.top.setXLink(self.bottom)
        else:
            self.top.setXLink(None)
            self.bottom.setXLink(None)

    def set_connect_y(self, connect: bool) -> None:
        if connect:
            self.top.setYLink(self.bottom)
        else:
            self.top.setYLink(None)
            self.bottom.setYLink(None)


class PlotWidget(StaticPlotWidget):
    """ A widget containing and handling a single plot. 
    
    Adds the look-and-feel for our omc3 guis to the default accwidgets plotwidget. 
    """
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs, viewBox=ZoomingViewBox())  # using viewbox here requires accwidgets >= 3.0.11
        self.setBackground("w")
        self._set_show_grid(ExPlotWidgetProperties.GridOrientationOptions.BOTH)


class ZoomingViewBox(ExViewBox):
    """ ViewBox that imitates the behavior of the Java-GUI a bit more closely than the default. """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setMouseMode(ZoomingViewBox.RectMode)  # mode that makes zooming rectangles

    # Look and Feel ------------------------------------------------------------
    def suggestPadding(self, axis):
        """ Suggests padding (between the data and the axis) for the autoRange function. 
        For our purposes, we do not want any padding on the x-axis. """
        if axis == 0:
            return 0.0  # disable padding for x axis
        return super().suggestPadding(axis)

    # Mouse Events -------------------------------------------------------------
    def mouseClickEvent(self, ev: mouseEvents.MouseClickEvent):
        if ev.button() == Qt.MouseButton.MiddleButton:
            ev.accept()
            self.auto_zoom()
            return 

        if ev.button() == Qt.MouseButton.RightButton:
            ev.accept()
            if ev.modifiers() == Qt.KeyboardModifier.AltModifier:
                self.raiseContextMenu(ev)
                return

            self.undo_zoom(reset=ev.modifiers() == Qt.KeyboardModifier.ShiftModifier)
            return

        super().mouseClickEvent(ev)
        
    def mouseDoubleClickEvent(self, ev: QGraphicsSceneMouseEvent):
        if ev.button() == Qt.MouseButton.LeftButton:
            ev.accept()
            self.undo_zoom(reset=ev.modifiers() == Qt.KeyboardModifier.ShiftModifier)
            return

        super().mouseDoubleClickEvent(ev)

    # Zoom History -------------------------------------------------------------
    def undo_zoom(self, reset: bool = False):
        """ Go back in zoom history. """
        if self.axHistoryPointer == 0:
            self.enableAutoRange()
            self.axHistoryPointer = -1
            self.save_view()
            return

        if reset:
            # Go back to the first zoom
            self.scaleHistory(-len(self.axHistory))
            return

        # go one step back
        self.scaleHistory(-1)

    def save_view(self):
        """ Save the current view to the zoom history. """
        self.axHistoryPointer += 1
        self.axHistory = self.axHistory[:self.axHistoryPointer] + [self.viewRect()]

    # Auto Zoom ----------------------------------------------------------------
    def auto_zoom(self):
        """ Zoom to 6, 4 and 2 standard deviations and save the steps. """
        for nsigma in (6, 4, 2):
            self.set_y_range_to_n_sigma(nsigma)
            self.save_view()

    def set_y_range_to_n_sigma(self, n_sigma):
        """ Set the y-range to a number of standard deviations,
        assuming the data is taken from a normal distribution. """
        # Get the data from all curves in the viewbox
        all_data = []
        for item in self.allChildren():
            if isinstance(item, pg.PlotDataItem):
                y_data = item.yData
                if y_data is not None:
                    all_data.extend(y_data)

        if not all_data:
            return

        all_data = np.array(all_data)
        mean = np.mean(all_data)
        std_dev = np.std(all_data)

        y_min = mean - n_sigma * std_dev
        y_max = mean + n_sigma * std_dev

        self.setYRange(y_min, y_max, padding=0)



"""
Settings
--------

Global Settings for the Segment-by-Segment application.
"""
from __future__ import annotations

from pathlib import Path
from dataclasses import dataclass, field

from omc3_gui.ui_components.dataclass_ui import metafield

@dataclass(slots=True)
class MainSettings:
    cwd: Path = metafield("Working Directory", "Current working directory. Used for default path when opening file selection dialogs.", default=Path.cwd())
    autoload_segments: bool = metafield("Autoload Segments", "Automatically try to load existing segments when loading a measurement.", default=True) 
    autodefault_segments: bool = metafield("Auto-Add Default Segments", "Automatically add default segments when loading a measurement.", default=False) 
    suggest_correctors: bool = metafield("Suggest Correctors", "Suggest correctors when editing a new correction file.", default=True)

@dataclass(slots=True)
class PlotSettings:
    show_model: bool = metafield("Show Model", "Show markers for the elements of the Model.", default=False)
    show_legend: bool = metafield("Show Legend", "Show legend.", default=True)
    marker_size: float = metafield("Marker Size", "Size of the markers.", default=8.5)
    expected: bool = metafield("Expectation", "Show expected value after correction instead of correction itself.", default=False)
    forward: bool = metafield("Forward Propagation", "Show forward propagation.", default=True)
    backward: bool = metafield("Backward Propagation", "Show backward propagation.", default=False)
    connect_x: bool = metafield("Connect X", "Connect X axes of the two plots.", default=True)
    connect_y: bool = metafield("Connect Y", "Connect Y axes of the two plots.", default=False)
    reset_zoom: bool = metafield("Reset Zoom", "Reset zoom when changing segments.", default=True)
    same_start: bool = metafield("Same Segment Start", "Plot only if the selected segments all have the same starting BPM.", default=True)
    model_s: bool = metafield("Model Location", "Use the model longitudinal location instead of the segment location.", default=False)


@dataclass(slots=True)
class Settings:
    main: MainSettings = field(default_factory=MainSettings)
    plotting: PlotSettings = field(default_factory=PlotSettings)




""" 
DataClass UI: Model 
-------------------

The backend components that make up the dataclass UI.
These are used as inputs to the controllers to define 
the meta-data and the type of the dataclass fields, 
so that the controller can find the fitting widgets and provide 
the user with additional information.
"""
from __future__ import annotations

import logging
from collections.abc import Callable
from dataclasses import MISSING, Field, dataclass, field
from pathlib import Path
from typing import Any

LOGGER = logging.getLogger(__name__)

@dataclass(slots=True)
class MetaData:
    """ Metadata for a dataclass-field. """
    label: str | None = None
    comment: str | None = None
    validate: Callable | None = None  # needs to return a truthy value if valid, else falsy or raise

    def __getitem__(self, key):
        return getattr(self, key)

    def get(self, key, default=None):
        return getattr(self, key, default)

    
def metafield(
    label: str, 
    comment: str, 
    default=MISSING, 
    validate: Callable | None = None, 
    ) -> Field:
    """ Convenience function to create a dataclass-field with metadata. """
    return field(
        default=default, 
        metadata=MetaData(label=label, comment=comment, validate=validate)
    )


def choices_validator(*choices: Any) -> Callable:
    """ Return a validator that checks if a given value is in the choices. """
    def validator(value):
        if value not in choices:
            raise ValueError(f"Value {value} is not in {choices}.")
        return True
    return validator


class FilePath(type(Path())):
    """ Convenience Class to indicate that the Path should lead to a file. 
    For the inheritance: https://stackoverflow.com/questions/29850801/subclass-pathlib-path-fails
    """
    pass


class DirectoryPath(type(Path())):
    """ Convenience Class to indicate that the Path should lead to a directory. 
    For the inheritance: https://stackoverflow.com/questions/29850801/subclass-pathlib-path-fails
    """
    pass


@dataclass(slots=True)
class FieldUIDef:
    """ Definition of an FieldUI to be generated by 
    :class:`omc3_gui.utils.dataclass_ui.controller.DataClassUI`.
    This defines how the label and edit-field widgets should look like
    for a field in the dataclass to be UI'ed.
    Missing information is parsed in from the dataclass 
    isetlf if possible, but values given here take precedence. 
    """
    name: str                        # name of the field in the dataclass
    label: str | None = None         # label of the field
    pytype: type | None = None       # type of the field's data, needs to be instanciable
    comment: str | None = None       # comment for the field, e.g. used for tooltips
    editable: bool | None = True     # sets field to be editable


""" 
UI: Threads
-----------

Helper functions for threads.
"""
import logging
from qtpy.QtCore import QThread, Signal
from collections.abc import Callable

LOGGER = logging.getLogger(__name__)


class BackgroundThread(QThread):
    """ Thread that runs a function in the background,
    providing a post- and exception-hook.
    
    Args:
        function (Callable): The function to run in the background.
        message (str): The message to display while the function is running.
        on_end_function (Callable): A function to call when the function is done.
        on_exception_function (Callable): A function to call when the function throws an exception.
    """

    on_exception = Signal([str])

    def __init__(self, 
                function: Callable,
                message: str = None,
                on_end_function: Callable = None, 
                on_exception_function: Callable = None):
        QThread.__init__(self)
        self._function = function
        self._message = message
        self._on_end_function = on_end_function
        self._on_exception_function = on_exception_function
    
    @property
    def message(self):
        return self._message

    def run(self):
        """ 
        Runner for the thread.
        This function is called automatically when the thread is started. 
        """
        try:
            self._function()
        except Exception as e:
            LOGGER.exception(str(e))
            self.on_exception.emit(str(e))
            return 
            
        LOGGER.info(f"Finished {self.message!s} successfully!")

    def start(self):
        """
        Connects the thread to the on_end and on_exception signals and starts the thread.
        This function is triggered manually to start the thread.
        """
        self.finished.connect(self._on_end)
        self.on_exception.connect(self._on_exception)
        super(BackgroundThread, self).start()

    def _on_end(self):
        if self._on_end_function:
            self._on_end_function()

    def _on_exception(self, exception_message):
        if self._on_exception_function:
            self._on_exception_function(exception_message)



""" 
Measurement View
----------------

This module contains the view for the measurement dialog.
"""
from dataclasses import fields

from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement
from omc3_gui.ui_components.dataclass_ui import DataClassDialog, FieldUIDef, DataClassUI



class OpticsMeasurementDialog(DataClassDialog):

    WINDOW_TITLE = "Optics Measurement"
    DEFAULT_SIZE = (800, -1)
    
    def __init__(self, parent=None, optics_measurement: OpticsMeasurement | None = None):
        if optics_measurement is None:
            optics_measurement = OpticsMeasurement()

        non_editable = ("measurement_dir", )  # set by program not by user
        dataclass_ui = DataClassUI(
            field_definitions=[
                FieldUIDef(field.name, editable=field.name not in non_editable) 
                for field in fields(OpticsMeasurement) if field.name[0] != "_"
            ],
            dclass=optics_measurement,
        )
        super().__init__(dataclass_ui=dataclass_ui, parent=parent)

    @property
    def measurement(self) -> OpticsMeasurement:
        return self._dataclass_ui.model


""" 
Defaults
--------

Defaults for segment by segment.
"""
from __future__ import annotations

from typing import TYPE_CHECKING
from omc3_gui.segment_by_segment.segment_model import SegmentTuple

if TYPE_CHECKING: 
    from omc3_gui.segment_by_segment.measurement_model import OpticsMeasurement

DEFAULT_SEGMENTS =(
    SegmentTuple("IP1", "BPM.12L1", "BPM.12R1"),
    SegmentTuple("IP2", "BPM.12L2", "BPM.12R2"),
    SegmentTuple("IP5", "BPM.12L5", "BPM.12R5"),
    SegmentTuple("IP8", "BPM.12L8", "BPM.12R8"),
)


LHC_ARCS = ('78', '81', '12', '23', '34', '45', '56', '67')  # 78 needs to be before 81 to make lr fit
LHC_CORRECTORS = (
    "kqf.a{}", 
    "kqd.a{}", 
    "kqt12.{}{}b{}", 
    "kqtl11.{}{}b{}", 
    "kq10.{}{}b{}",
    "kq9.{}{}b{}",
    "kq8.{}{}b{}",
    "kq7.{}{}b{}",
    "kq6.{}{}b{}",
    "kq5.{}{}b{}",
    "kq4.{}{}b{}",
    "kqsx3.{}{}",
    "ktqx2.{}{}",
    "ktqx1.{}{}",
    "kqx.{}{}",
)

def get_default_correctors(measurement: OpticsMeasurement) -> str:
    text = ""
    if measurement.accel == "lhc":
        text = ""
        for ip in (1, 2, 5, 8):
            text += f"! IP{ip} -----\n"
            arcs = [arc for arc in LHC_ARCS if str(ip) in arc]
            for arc, side in zip(arcs, "lr"):
                lhc_correctors = LHC_CORRECTORS if side == "l" else LHC_CORRECTORS[::-1]
                for corrector in lhc_correctors:
                    if "a" in corrector:
                        corrector = corrector.format(arc)
                    else:
                        corrector = corrector.format(side, ip, measurement.beam)
                    text += f"! {corrector} = {corrector};\n"
            text += "\n"
    return text 



"""
The omc3-gui library
~~~~~~~~~~~~~~~~~~~~

omc3-gui provides graphical user interfaces for the ``omc3`` package for the optics measurements and corrections group (OMC) at CERN.

:copyright: pyLHC/OMC-Team working group.
"""

__title__ = "omc3-gui"
__description__ = "QT Graphical User Interface wrapper of the ``omc3`` package"
__url__ = "https://github.com/pylhc/omc3_gui"
__version__ = "1.0.1"
__author__ = "pylhc"
__author_email__ = "pylhc@github.com"

__all__ = [__version__]



""" 
Segment View
------------

This module contains the view for the segment dialog.
"""
from __future__ import annotations

from omc3_gui.segment_by_segment.segment_model import SegmentDataModel
from omc3_gui.ui_components.dataclass_ui import DataClassDialog, DataClassUI, FieldUIDef

class SegmentDialog(DataClassDialog):

    WINDOW_TITLE = "Segment Editor"
    DEFAULT_SIZE = (400, -1)
    
    def __init__(self, parent=None, segment: SegmentDataModel | None = None):
        if segment is None:
            segment = SegmentDataModel(measurement=None)  # dummy
        
        dataclass_ui = DataClassUI(
            field_definitions=[
                FieldUIDef(name) for name in ("name", "start", "end")
            ],
            dclass=segment,
        )
        super().__init__(dataclass_ui=dataclass_ui, parent=parent)
    
    @property
    def segment(self) -> SegmentDataModel:
        return self._dataclass_ui.model



""" 
Measurement Model
-----------------

This module contains the model for the Optics Measurement 
in the Segment-by-Segment application.
"""
from __future__ import annotations

import json
import logging
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from omc3.model.constants import TWISS_DAT
from omc3.optics_measurements.constants import (
    BETA_NAME,
    EXT,
    KICK_NAME,
    MODEL_DIRECTORY,
    PHASE_NAME,
)
from omc3.segment_by_segment.constants import corrections_madx
from tfs.reader import read_headers

from omc3_gui.ui_components.dataclass_ui import DirectoryPath, FilePath, metafield
from omc3_gui.ui_components.dataclass_ui import choices_validator as choices
from omc3_gui.ui_components.dataclass_ui.tools import (
    load_dataclass_from_json,
    save_dataclass_to_json,
    update_dataclass_from_json,
)

if TYPE_CHECKING:
    from omc3_gui.segment_by_segment.segment_model import SegmentDataModel

SEQUENCE: str = "SEQUENCE"
DATE: str = "DATE"

FILES_TO_LOOK_FOR: tuple[str, ...] = tuple(f"{name}{plane}" for name in (KICK_NAME, PHASE_NAME, BETA_NAME) for plane in ("x", "y"))
TO_BE_DEFINED: str = "to_be_defined"


LOGGER = logging.getLogger(__name__)


def exists(value: Path | None) -> bool:
    return value is not None and value.exists()


@dataclass(slots=True)
class OpticsMeasurement:
    """ Class to load and hold the optics-measurement folder. 
    This class also stores the meta-data for the loaded measurement, 
    which can then be passed on to the segment-by-segment.
    The :func:`omc3_gui.utils.dataclass_ui.metafield` is used to provide hints about the fields for the GUI.
    """
    measurement_dir: DirectoryPath = metafield("Optics Measurement", "Path to the optics-measurement folder", default=Path(TO_BE_DEFINED), validate=exists)
    model_dir: DirectoryPath =       metafield("Model",              "Path to the model folder",        default=Path(TO_BE_DEFINED), validate=exists)
    accel: str =                     metafield("Accelerator",        "Name of the accelerator",         default=None)
    output_dir: DirectoryPath =      metafield("Output",             "Path to the sbs-output folder",   default=None) 
    corrections: FilePath =          metafield("Corrections",        "Path to the corrections file",    default=None)
    year: str =                      metafield("Year",               "Year of the measurement (model)", default=None)
    ring: int =                      metafield("Ring",               "Ring of the accelerator",         default=None, validate=choices(1, 2, 3, 4))
    beam: int =                      metafield("Beam",               "Beam of the accelerator",         default=None, validate=choices(1, 2)) 
    # List of segments. Using a list here, so the name and start/end can be changed
    # without having to modify anything here.
    _segments: list[SegmentDataModel] = field(default_factory=list)

    DEFAULT_OUTPUT_DIR: ClassVar[str] = "sbs"
    JSON_FILENAME: ClassVar[str] = "sbs_measurement.json"

    def __post_init__(self):
        if self.output_dir is None:
            self.output_dir = self.measurement_dir / self.DEFAULT_OUTPUT_DIR

    # Visualization ------------------------------------------------------------
    def display(self) -> str:
        if self.output_dir.name == self.DEFAULT_OUTPUT_DIR:
            return self.measurement_dir.name
        return f"{self.measurement_dir.name} -> {self.output_dir.name}"

    @property
    def id(self) -> str:
        """ Unique identifier for the measurement, used in the ItemModel. """
        return str(self.output_dir)

    @classmethod
    def get_label(cls, name: str) -> str:
        """ Returns the label for the field named `name`. """
        try:
            return cls.__dataclass_fields__[name].metadata["label"]
        except KeyError:
            return name

    @classmethod
    def get_comment(cls, name: str) -> str:
        """ Returns the comment for the field named `name`. """
        try:
            return cls.__dataclass_fields__[name].metadata["comment"]
        except KeyError:
            return ""

    def tooltip(self) -> str:
        """ Returns a string with information about the measurement, 
        as to be used in a tool-tip.  """
        parts = [
            (self.get_label(f.name), getattr(self, f.name)) for f in fields(self) 
            if not f.name.startswith("_")
        ]
        size = max(len(name) for name, _ in parts)
        return "\n".join(f"{name:{size}s}: {value}" for name, value in parts if value is not None)

    # Segment Control ----------------------------------------------------------
    def remove_segment(self, segment: SegmentDataModel):
        try:
            self.segments.remove(segment)
        except ValueError as e:
            raise ValueError(f"Segment with name {segment.name} is not in {self.display()}.") from e

    def add_segment(self, segment: SegmentDataModel):
        if segment in self.segments:
            raise NameError(f"Segment {segment} is already in {self.display()}")
        
        if segment.name in [s.name for s in self.segments]:
            raise NameError(f"A segment with name {segment.name} is already in {self.display()}")

        self.segments.append(segment)
    
    def try_add_segment(self, segment: SegmentDataModel, silent: bool = False) -> bool:
        try:
            self.add_segment(segment)
        except NameError as e:
            if silent:
                LOGGER.debug(str(e))
            else: 
                LOGGER.error(str(e))
            return False
        return True
    
    def try_remove_segment(self, segment: SegmentDataModel | str) -> bool:
        if isinstance(segment, str):
            try:
                segment = self.get_segment_by_name(segment)
            except NameError as e:
                LOGGER.error(str(e))
                return False

        try:
            self.remove_segment(segment)
        except ValueError as e:
            LOGGER.error(str(e))
            return False
        return True
    
    def get_segment_by_name(self, name: str) -> SegmentDataModel:
        for segment in self.segments:
            if segment.name == name:
                return segment

        msg = f"No segment with name {name} in {self.display()}."
        raise NameError(msg)
        
    @property
    def segments(self) -> list[SegmentDataModel]:
        return self._segments

    # Segment-by-Segment Parameters --------------------------------------------
    def get_sbs_parameters(self) -> dict[str, Any]:
        parameters = dict(
            measurement_dir=self.measurement_dir,
            corrections=self.corrections,
            output_dir=self.output_dir,
            accel = self.accel,
            model_dir = self.model_dir,
        )
        if self.beam is not None:
            parameters["beam"] = self.beam
        if self.year is not None:
            parameters["year"] = self.year
        if self.ring is not None:
            parameters["ring"] = self.ring 
        return parameters

    # Builder ------------------------------------------------------------------
    def copy(self) -> OpticsMeasurement:
        """ Creates a copy of the measurement. """
        new_measurement = OpticsMeasurement(
            **{
                f.name: getattr(self, f.name) for f in fields(self) 
                if not f.name.startswith("_")
            },
        )
        for segment in self.segments:
            new_segment = segment.copy()
            new_segment.measurement = new_measurement
            new_measurement.add_segment(new_segment)
        return new_measurement

    @classmethod
    def from_path(cls, path: Path) -> OpticsMeasurement:
        """ Creates an OpticsMeasurement from a folder, by trying 
        to parse information from the data in the folder.

        Args:
            path (Path): Path to the folder.

        Returns:
            OpticsMeasurement: OpticsMeasurement instance. 
        """
        # Try to load from json first ---
        json_path = None
        default_json_path = path / cls.JSON_FILENAME
        default_output_json_path = path / cls.DEFAULT_OUTPUT_DIR / cls.JSON_FILENAME
        measurement_path = None
        if (default_json_path).is_file():
            json_path = default_json_path
            if any((path / f).is_file() for f in FILES_TO_LOOK_FOR):
                measurement_path = path  # otherwise probably an output dir
        elif(default_output_json_path).is_file():
            json_path = default_output_json_path
            measurement_path = path
        
        if json_path is not None:
            try:
                return cls.from_json(json_path, measurement_path)
            except json.decoder.JSONDecodeError as e:
                LOGGER.error(f"JSON errror: {e!s}\nTrying to load as optics-measurement folder.")

        # Try to load from optics-measurement folder ---
        info = {}
        try:
            model_dir = _parse_model_dir_from_optics_measurement(path)
        except FileNotFoundError as e:
            LOGGER.error(str(e))
        else:
            info = _parse_info_from_model_dir(model_dir)
            info["model_dir"] = model_dir
        
        if (path / corrections_madx).is_file():
            info["corrections"] = path / corrections_madx
        
        return cls(measurement_dir=path, **info)
    
    @classmethod
    def from_json(cls, path: Path, measurement_dir: Path | None = None) -> OpticsMeasurement:
        """ Creates an OpticsMeasurement from a folder, by trying 
        to parse information from the data in the folder.

        Args:
            path (Path): Path to the folder.

        Returns:
            OpticsMeasurement: OpticsMeasurement instance. 
        """
        if measurement_dir is not None:
            meas: OpticsMeasurement = cls(measurement_dir=measurement_dir)
            meas = update_dataclass_from_json(meas, path)
            meas.measurement_dir = measurement_dir  # in case the folder name/path changed since the json was written
            return meas
        return load_dataclass_from_json(cls, path)

    def to_json(self, path: Path | None = None) -> None:
        if path is None:
            if self.output_dir is not None:
                self.output_dir.mkdir(parents=True, exist_ok=True)
                path = self.output_dir / self.JSON_FILENAME
            else:
                if self.measurement_dir is None or self.measurement_dir == TO_BE_DEFINED:
                    raise ValueError("Measurement dir is still to be defined.")
                path = self.measurement_dir / self.JSON_FILENAME
        save_dataclass_to_json(self, path)
    
    def quick_check(self) -> None:
        """ Tests for completeness of the definition (e.g. after loading). """
        if self.measurement_dir is None or self.measurement_dir == TO_BE_DEFINED:
            raise NameError("Measurement dir is still to be defined.")  # BAD

        if any(getattr(self, name) is None for name in ("model_dir", "accel", "output_dir")) or self.model_dir == TO_BE_DEFINED:
            raise ValueError(f"Current definition of '{self.measurement_dir!s}' is incomplete. Adjust manually!!")

        if self.accel == 'lhc' and (self.year is None or self.beam is None):
            raise ValueError(f"Current definition of '{self.measurement_dir!s}' for LHC is incomplete. Adjust manually!!")        
        
        if self.accel == 'psb' and self.ring is None:
            raise ValueError(f"Current definition of '{self.measurement_dir!s}' for PSB is incomplete. Adjust manually!!")


def _parse_model_dir_from_optics_measurement(measurement_path: Path) -> Path:
    """Tries to find the model directory in the headers of one of the optics measurement files.

    Args:
        measurement_path (Path): Path to the folder. 

    Returns:
        Path: Path to the (associated) model directory. 
    """
    LOGGER.debug(f"Searching for model dir in {measurement_path!s}")
    for file_name in FILES_TO_LOOK_FOR:
        LOGGER.debug(f"Checking {file_name!s} for model dir.")
        try:
            headers = read_headers((measurement_path / file_name).with_suffix(EXT))
        except FileNotFoundError:
            LOGGER.debug(f"{file_name!s} not found in {measurement_path!s}.")
        else:
            if MODEL_DIRECTORY in headers:
                LOGGER.debug(f"{MODEL_DIRECTORY!s} found in {file_name!s}: {headers[MODEL_DIRECTORY]!s}!")
                return Path(headers[MODEL_DIRECTORY])

            LOGGER.debug(f"{MODEL_DIRECTORY!s} not found in {file_name!s}.")
    raise FileNotFoundError(f"Could not find '{MODEL_DIRECTORY}' in any of {FILES_TO_LOOK_FOR!r} in {measurement_path!r}")


def _parse_info_from_model_dir(model_dir: Path) -> dict[str, Any]:
    """ Checking twiss.dat for more info about the accelerator.

    Args:
        model_dir (Path): Path to the model-directory. 

    Returns:
        Dict[str, Any]: Containing the additional info found (accel, beam, year, ring). 
    """
    result = {}

    try:
        headers = read_headers(model_dir / TWISS_DAT)
    except FileNotFoundError as e:
        LOGGER.debug(str(e))
        return result

    sequence = headers.get(SEQUENCE)
    if sequence is not None:
        sequence = sequence.lower()
        if "lhc" in sequence:
            result['accel'] = "lhc"
            result['beam'] = int(sequence[-1])
            result['year'] = map_lhc_year(_get_year_from_header(headers))
        elif "psb" in sequence:
            result['accel'] = "psb"
            result['ring'] = int(sequence[-1])
        else:
            result['accel'] = sequence
    LOGGER.debug(f"Associated info found in model dir '{model_dir!s}':\n {result!s}")
    return result


def _get_year_from_header(headers: dict) -> str | None:
    """ Parses the year from the date in the twiss.dat file.
    
    TODO: Will not work for hl-lhc models. These should return the hl-version.
    """
    date = headers.get(DATE)
    
    if date is None:
        return None

    year = f"20{date.split('/')[-1]}"
    LOGGER.debug(f"Assuming model year {year!s} from '{date}'!")
    return year


def map_lhc_year(year: str | None) -> str:
    """ Maps the input year to the corresponding available model year. """
    if year is None:
        return None
    
    try:
        int_year = int(year)
    except ValueError:
        return year
    
    # no new models (see omc/model/accelerators/lhc)
    if 2012 < int_year < 2015:  
        LOGGER.info(f"Mapping year {year} to LHC model 2012!")
        return "2012"
    
    # no new models (there was a 2021 in acc-models, but we were not using it then)
    if 2018 < int_year < 2022:  
        LOGGER.info(f"Mapping year {year} to LHC model 2018!")
        return "2018"
    
    return year
    